//
// SMPBasicBlock.cpp
//
// This module performs the fundamental basic block level analyses needed for the
//   SMP project (Software Memory Protection).
//

#include <list>
#include <set>
#include <vector>
#include <algorithm>

#include <cstring>

#include <pro.h>
#include <assert.h>
#include <ida.hpp>
#include <idp.hpp>
#include <auto.hpp>
#include <bytes.hpp>
#include <funcs.hpp>
#include <name.hpp>
#include <allins.hpp>
#include <intel.hpp>

#include "SMPDataFlowAnalysis.h"
#include "SMPBasicBlock.h"
#include "SMPInstr.h"
#include "SMPFunction.h"

#define SMP_DEBUG_DATAFLOW 0
#define SMP_DEBUG_OPTIMIZATIONS 0

// Basic block number 0 is the top of the CFG lattice.
#define SMP_TOP_BLOCK 0 

// *****************************************************************
// Class SMPBasicBlock
// *****************************************************************

// Constructor
SMPBasicBlock::SMPBasicBlock(SMPFunction *Func, list<SMPInstr>::iterator First, list<SMPInstr>::iterator Last) {
	this->FirstAddr = First->GetAddr();
	this->BlockNum = SMP_BLOCKNUM_UNINIT;
	this->MyFunc = Func;
	this->Processed = false;
	this->IndirectJump = false;
	this->Returns = false;
	this->SharedTailChunk = false;
	this->IsLiveInStale = true;
	this->FlagsDeadAfterLastUse = false;
	this->FlagsDeadBeforeFirstKill = false;

	this->FirstFlagsDef = Last->GetAddr() + 1;
	this->LastFlagsUse = First->GetAddr() - 1;

	this->Predecessors.clear();
	this->Successors.clear();
	this->KillSet.clear();
	this->UpExposedSet.clear();
	this->LiveOutSet.clear();
	this->LiveInSet.clear();
	this->DomFrontier.clear();
	this->PhiFunctions.clear();
	this->LocalNames.clear();

	list<SMPInstr>::iterator CurrInst = First;
	while (CurrInst != Last) {
		this->Instrs.push_back(CurrInst);
		if (CurrInst->HasFlagsDef() && (CurrInst->GetAddr() < this->FirstFlagsDef))
			this->FirstFlagsDef = CurrInst->GetAddr();
		if (CurrInst->HasFlagsUse() && (CurrInst->GetAddr() > this->LastFlagsUse))
			this->LastFlagsUse = CurrInst->GetAddr();
		++CurrInst;
	}

	// Now process the last instruction.
	if (Last->HasFlagsDef() && (Last->GetAddr() < this->FirstFlagsDef))
		this->FirstFlagsDef = Last->GetAddr();
	if (Last->HasFlagsUse() && (Last->GetAddr() > this->LastFlagsUse))
		this->LastFlagsUse = Last->GetAddr();
	Last->SetTerminatesBlock();
	this->Instrs.push_back(Last); // Add last instruction
}

// Get address of first instruction in the block.
ea_t SMPBasicBlock::GetFirstAddr(void) const {
	return this->FirstAddr;
}

bool SMPBasicBlock::IsLocalName(op_t CurrOp) const {
	return (!this->LocalNames.empty()
		&& (this->LocalNames.end() != this->LocalNames.find(CurrOp)));
}

set<op_t, LessOp>::iterator SMPBasicBlock::FindLocalName(op_t FindOp) {
	return this->LocalNames.find(FindOp);
}

// Equality operator for SMPBasicBlock. Key field is address of first instruction.
int SMPBasicBlock::operator==(const SMPBasicBlock &rhs) const {
	if (rhs.GetFirstAddr() != this->FirstAddr)
		return 0;
	else
		return 1;
}

// Link to predecessor block.
void SMPBasicBlock::LinkToPred(list<SMPBasicBlock>::iterator Predecessor) {
	this->Predecessors.push_back(Predecessor);
	return;
}

// Link to successor block.
void SMPBasicBlock::LinkToSucc(list<SMPBasicBlock>::iterator Successor) {
	this->Successors.push_back(Successor);
	return;
}

// See if all predecessors have set their ordering number.
bool SMPBasicBlock::AllPredecessorsNumbered(void) {
	list<list<SMPBasicBlock>::iterator>::iterator CurrPred;
	for (CurrPred = this->Predecessors.begin(); CurrPred != this->Predecessors.end(); ++CurrPred) {
		// Don't count current block, in case we have a one-block loop with this block
		//  as its own predecessor.
		if (**CurrPred == *this)
			continue;
		if ((*CurrPred)->GetNumber() == SMP_BLOCKNUM_UNINIT)
			return false;
	}
	return true;
} // end of SMPBasicBlock::AllPredecessorsNumbered()

// Are all instructions in the block no-ops?
bool SMPBasicBlock::AllNops(void) {
	size_t NopCount = 0;
	size_t GoodCount = 0;  // non-nop instructions
	list<list<SMPInstr>::iterator>::iterator CurrInst;
	for (CurrInst = this->Instrs.begin(); CurrInst != this->Instrs.end(); ++CurrInst) {
		if ((*CurrInst)->MDIsNop())
			++NopCount;
		else
			++GoodCount;
	}
	return ((0 == GoodCount) && (0 < NopCount));
} // end of SMPBasicBlock::AllNops()

// Analyze basic block and fill data members.
void SMPBasicBlock::Analyze() {
	if (Instrs.back()->GetDataFlowType() == INDIR_JUMP) {
		this->IndirectJump = true;
	}
	else if (Instrs.back()->GetDataFlowType() == RETURN) {
		this->Returns = true;
	}
} // end of SMPBasicBlock::Analyze()

// DEBUG dump of block
void SMPBasicBlock::Dump(void) {
	msg("Dump of basic block %d\n", this->BlockNum);
	// Dump dataflow analysis sets and links before dumping instructions.
	list<list<SMPBasicBlock>::iterator>::iterator CurrLink;
	msg("Predecessors: ");
	if (!this->Predecessors.empty()) {
		for (CurrLink = this->Predecessors.begin(); CurrLink != this->Predecessors.end(); ++CurrLink) {
			msg("%d ", (*CurrLink)->GetNumber());
		}
	}
	msg("\n");
	msg("Successors: ");
	if (!this->Successors.empty()) {
		for (CurrLink = this->Successors.begin(); CurrLink != this->Successors.end(); ++CurrLink) {
			msg("%d ", (*CurrLink)->GetNumber());
		}
	}
	msg("\n");
	set<op_t, LessOp>::iterator SetItem;
	msg("VarKill set: ");
	if (!this->KillSet.empty()) {
		for (SetItem = this->KillSet.begin(); SetItem != this->KillSet.end(); ++SetItem) {
			PrintListOperand(*SetItem);
		}
	}
	msg("\n\n");
	msg("UpExposed set: ");
	if (!this->UpExposedSet.empty()) {
		for (SetItem = this->UpExposedSet.begin(); SetItem != this->UpExposedSet.end(); ++SetItem) {
			PrintListOperand(*SetItem);
		}
	}
	msg("\n\n");
	msg("LiveIn set: ");
	if (!this->LiveInSet.empty()) {
		for (SetItem = this->LiveInSet.begin(); SetItem != this->LiveInSet.end(); ++SetItem) {
			PrintListOperand(*SetItem);
		}
	}
	msg("\n\n");
	msg("LiveOut set: ");
	if (!this->LiveOutSet.empty()) {
		for (SetItem = this->LiveOutSet.begin(); SetItem != this->LiveOutSet.end(); ++SetItem) {
			PrintListOperand(*SetItem);
		}
	}
	msg("\n\n");
	msg("Dominance frontier: ");
	set<int>::iterator DomIter;
	if (!this->DomFrontier.empty()) {
		for (DomIter = this->DomFrontier.begin(); DomIter != this->DomFrontier.end(); ++DomIter) {
			msg("%d ", *DomIter);
		}
	}
	msg("\n\n");
	set<SMPPhiFunction, LessPhi>::iterator PhiIter;
	if (!this->PhiFunctions.empty()) {
		for (PhiIter = this->PhiFunctions.begin(); PhiIter != this->PhiFunctions.end(); ++PhiIter) {
			msg("Phi function for %d : ", PhiIter->GetIndex());
			// Dump out all phi operands
			PhiIter->Dump();
		}
	}
	msg("\n");

	msg("DEF-USE chains for block-local names: \n");
	this->LocalDUChains.Dump();

#if SMP_FULL_LIVENESS_ANALYSIS
	msg("\n");
	msg("DEF-USE chains for block-local use of global names: \n");
	this->GlobalDUChains.Dump();
#endif

	msg("\n");
	if (this->IndirectJump)
		msg("Has indirect jump. ");
	if (this->Returns)
		msg("Has return. ");
	if (this->SharedTailChunk)
		msg("Is shared tail chunk block. ");
	msg("\n");

	// Now, dump all the instructions.
	list<list<SMPInstr>::iterator>::iterator CurrInst;
	for (CurrInst = this->Instrs.begin(); CurrInst != this->Instrs.end(); ++CurrInst) {
		(*CurrInst)->Dump();
	}
	msg("\n");
	return;
} // end of SMPBasicBlock::Dump()

// Return true if anything already in the KillSet would kill the operand value.
bool SMPBasicBlock::MDAlreadyKilled(op_t Opnd1) const {
	// We have assembly language operands that can be complex, such as
	//  [ebx + esi*4 + 04h]. If ebx or esi have been killed, then this memory
	//  phrase should be considered killed. We could be even more conservative
	//  with base addresses, declaring an entire array killed whenever its base
	//  address appears in a definition, for example. We will do that if it proves
	//  to be necessary.
	bool FoundInKillSet = (KillSet.end() != KillSet.find(Opnd1));
	switch (Opnd1.type) {
		// Some types are simple to test for equality.
		case o_void:
		case o_reg:
		case o_mem:
		case o_imm:
		case o_far:
		case o_near:
		case o_trreg:
		case o_dbreg:
		case o_crreg:
		case o_fpreg:
		case o_mmxreg:
		case o_xmmreg:
			// Look in KillSet. These simple types should be found there with
			//  no complicated comparisons needed.
			return FoundInKillSet;

		case o_phrase:
		case o_displ:
			// If found directly in KillSet, return true. Otherwise, see if any registers
			//  used in the memory addressing expression were killed.
			if (FoundInKillSet)
				return true;
			else {
				// Should we add Opnd1 to the KillSet every time we return true below? **!!**
				op_t TempOp;
				TempOp.type = o_reg;
				int BaseReg;
				int IndexReg;
				ushort ScaleFactor;
				ea_t displacement;

				MDExtractAddressFields(Opnd1, BaseReg, IndexReg, ScaleFactor, displacement);
			
				if (R_none != BaseReg) {
					TempOp.reg = (ushort) BaseReg;
					if (this->KillSet.end() != this->KillSet.find(TempOp))
						return true;
				}
				if (R_none != IndexReg) { // Cannot have ESP index reg in SIB
					TempOp.reg = (ushort) IndexReg;
					if (this->KillSet.end() != this->KillSet.find(TempOp))
						return true;
				}
			} // end if (FoundInKillSet) ... else ...
			break;
		default:
			msg("Unknown operand type %d in MDAlreadyKilled, block %d\n", Opnd1.type, this->BlockNum);
	} // end of switch on Opnd1.type

	return false;
} // end of SMPBasicBlock::MDAlreadyKilled()

// Initialize the KilledSet and UpExposedSet for live variable analysis.
void SMPBasicBlock::InitKilledExposed(void) {
	// Find all upwardly exposed operands and killed operands in this block.
	list<list<SMPInstr>::iterator>::iterator CurrIter;
	for (CurrIter = this->Instrs.begin(); CurrIter != this->Instrs.end(); ++CurrIter) {
		list<SMPInstr>::iterator CurrInst = *CurrIter;

		// Dataflow equation for upward exposed variables: If a variable has not been
		//  killed yet in this block, starting from the top of the block, and it is used
		//  in the current instruction, then it is upwardly exposed.
		set<DefOrUse, LessDefUse>::iterator CurrUse;
		for (CurrUse = CurrInst->GetFirstUse(); CurrUse != CurrInst->GetLastUse(); ++CurrUse) {
			op_t UseOp = CurrUse->GetOp();
			// Only add non-immediate operands that are not already killed in this block.
			//  o_near and o_far operands are code addresses in immediate form, e.g.
			//  call _printf might be call 0x8048040, with o_near = 0x8048040.
			if ((!(this->MDAlreadyKilled(UseOp)))
				&& (UseOp.type != o_imm) && (UseOp.type != o_near) && (UseOp.type != o_far))
				this->UpExposedSet.insert(UseOp);
		}
		// Dataflow equation for killed variables: If a variable is defined in any
		//  instruction in the block, it is killed by this block (i.e. prior definitions
		//  of that variable will not make it through the block).
		set<DefOrUse, LessDefUse>::iterator CurrDef;
		for (CurrDef = CurrInst->GetFirstDef(); CurrDef != CurrInst->GetLastDef(); ++CurrDef) {
			this->KillSet.insert(CurrDef->GetOp());
		}
	} // end for all instrs in block
	this->IsLiveInStale = true;  // Would need to compute LiveInSet for first time
	return;
} // end of SMPBasicBlock::InitKilledExposed()

// Return an iterator for the beginning of the LiveInSet. If the set is stale,
//  recompute it first.
set<op_t, LessOp>::iterator SMPBasicBlock::GetFirstLiveIn(void) {
	if (this->IsLiveInStale) {
		// Dataflow equation: A variable is live-in to this block if it
		//  is upwardly exposed from this block, or if it passes through
		//  the block unchanged (i.e. it is not killed and is live out).
		this->LiveInSet.clear();
		set<op_t, LessOp>::iterator OutIter;
		for (OutIter = this->UpExposedSet.begin(); OutIter != this->UpExposedSet.end(); ++OutIter) {
			this->LiveInSet.insert(*OutIter);
		}
		for (OutIter = this->LiveOutSet.begin(); OutIter != this->LiveOutSet.end(); ++OutIter) {
			if (KillSet.end() == this->KillSet.find(*OutIter)) // Found live out but not killed
				this->LiveInSet.insert(*OutIter);
		}
		this->IsLiveInStale = false;
	}
	return this->LiveInSet.begin();
} // end of SMPBasicBlock::GetFirstLiveIn()

// Get termination iterator marker for the LiveIn set, for use by predecessors.
set<op_t, LessOp>::iterator SMPBasicBlock::GetLastLiveIn(void) {
	// Does not matter if it is stale or not; end marker is the same
	return this->LiveInSet.end();
}

// Get iterator for the start of the LiveOut set.
set<op_t, LessOp>::iterator SMPBasicBlock::GetFirstLiveOut(void) {
	return this->LiveOutSet.begin();
}

// Get termination iterator marker for the LiveOut set.
set<op_t, LessOp>::iterator SMPBasicBlock::GetLastLiveOut(void) {
	return this->LiveOutSet.end();
}

// Get iterator for the start of the VarKill set.
set<op_t, LessOp>::iterator SMPBasicBlock::GetFirstVarKill(void) {
	return this->KillSet.begin();
}

// Get termination iterator marker for the VarKill set.
set<op_t, LessOp>::iterator SMPBasicBlock::GetLastVarKill(void) {
	return this->KillSet.end();
}

// Get iterator for the start of the UpExposed set.
set<op_t, LessOp>::iterator SMPBasicBlock::GetFirstUpExposed(void) {
	return this->UpExposedSet.begin();
}

// Get termination iterator marker for the UpExposed set.
set<op_t, LessOp>::iterator SMPBasicBlock::GetLastUpExposed(void) {
	return this->UpExposedSet.end();
}

// Get iterator for the start of the LocalNames set.
set<op_t, LessOp>::iterator SMPBasicBlock::GetFirstLocalName(void) {
	return this->LocalNames.begin();
}

// Get termination iterator marker for the LocalNames set.
set<op_t, LessOp>::iterator SMPBasicBlock::GetLastLocalName(void) {
	return this->LocalNames.end();
}

// Get iterator for the start of the DomFrontier set.
set<int>::iterator SMPBasicBlock::GetFirstDomFrontier(void) {
	return this->DomFrontier.begin();
}

// Get termination iterator marker for the DomFrontier set.
set<int>::iterator SMPBasicBlock::GetLastDomFrontier(void) {
	return this->DomFrontier.end();
}

// Get iterator for first Phi function.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::GetFirstPhi(void) {
	return this->PhiFunctions.begin();
}

// Get termination iterator marker for Phi functions set.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::GetLastPhi(void) {
	return this->PhiFunctions.end();
}

// Find the Phi function that matches the given op_t.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::FindPhi(op_t FindOp) {
	set<SMPPhiFunction, LessPhi>::iterator PhiIter;
	set<op_t, LessOp>::value_compare OpComp = this->LiveOutSet.value_comp(); // LessOp()
	for (PhiIter = this->PhiFunctions.begin(); PhiIter != this->PhiFunctions.end(); ++PhiIter) {
		op_t PhiOp = PhiIter->GetPhiRef(0).GetOp();
		if (OpComp(FindOp, PhiOp) || OpComp(PhiOp, FindOp)) {
			// One is less than the other.
			; // do nothing
		}
		else {
			// They are equal.
			return PhiIter;
		}
	}
	// If we exit the loop, we did not find it.
	return this->PhiFunctions.end();
} // end of SMPBasicBlock::FindPhi()

// The predecessor order is used in SSA phi functions. Get the ordinal number
//  of a given block number in the predecessor list.
int SMPBasicBlock::GetPredPosition(int BlockNum) {
	int counter = 0;
	list<list<SMPBasicBlock>::iterator>::iterator PredIter;
	for (PredIter = this->GetFirstPred(); PredIter != this->GetLastPred(); ++PredIter) {
		if (BlockNum == (*PredIter)->GetNumber()) {
			return counter;
		}
		++counter;
	}
	return -1;
}

// Update the LiveOut set for the block.
// Return true if it changed, false otherwise.
bool SMPBasicBlock::UpdateLiveOut(void) {
	bool changed = false;
	set<op_t, LessOp> OldLiveOut(this->LiveOutSet); // save copy of old LiveOutSet
	this->LiveOutSet.clear();  // Clear it and rebuild it
	// Dataflow equation for LiveOutSet: If a variable is live-in for any successor
	//  block, it is live out for this block.
	list<list<SMPBasicBlock>::iterator>::iterator SuccIter;
	for (SuccIter = this->Successors.begin(); SuccIter != this->Successors.end(); ++SuccIter) {
		set<op_t, LessOp>::iterator InSuccIter;
		for (InSuccIter = (*SuccIter)->GetFirstLiveIn(); InSuccIter != (*SuccIter)->GetLastLiveIn(); ++InSuccIter) {
			this->LiveOutSet.insert(*InSuccIter);
		}
	}

	// Only remaining question: Did the LiveOutSet change?
	// Short cut: If the set cardinality changed, then the set changed.
	if (this->LiveOutSet.size() != OldLiveOut.size()) {
		changed = true;
	}
	else { // Same # of elements; move through in lockstep and compare.
		set<op_t, LessOp>::iterator NewIter = this->LiveOutSet.begin();
		set<op_t, LessOp>::iterator OldIter = OldLiveOut.begin();
		set<op_t, LessOp>::value_compare OpComp = OldLiveOut.value_comp(); // LessOp()
		while (OldIter != OldLiveOut.end()) { // both iters terminate simultaneously
			if (OpComp(*OldIter, *NewIter) || OpComp(*NewIter, *OldIter)) {
				changed = true;
				break;
			}
			++OldIter;
			++NewIter;
		}
	}

	if (changed)
		this->IsLiveInStale = true;

	OldLiveOut.clear();
	return changed;
} // end of SMPBasicBlock::UpdateLiveOut()

// Insert RPO number block into the dominance frontier set.
void SMPBasicBlock::AddToDomFrontier(int block) {
	this->DomFrontier.insert(block);
	return;
} // end of SMPBasicBlock::AddToDomFrontier()

// Fill the set of non-global names used in this block. Set local indices that can be
//  used later to index into a vector of def-use chains.
void SMPBasicBlock::SetLocalNames(void) {
	list<list<SMPInstr>::iterator>::iterator InstIter;
	size_t LocalIndex = 0;
#if SMP_DEBUG_DATAFLOW
	msg("Entered SetLocalNames.\n");
#endif
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		set<DefOrUse, LessDefUse>::iterator CurrDef;
		for (CurrDef = (*InstIter)->GetFirstDef(); CurrDef != (*InstIter)->GetLastDef(); ++CurrDef) {
			op_t DefOp = CurrDef->GetOp();
			if (!(this->MyFunc->IsGlobalName(DefOp))) {
				// Not in global names set
				if (this->LocalNames.end() == this->LocalNames.find(DefOp)) {
					// Not yet in LocalNames, so add it.
					SetGlobalIndex(&DefOp, LocalIndex);
					this->LocalNames.insert(DefOp);
					++LocalIndex;
				}
			}
		}
	}
	return;
} // end of SMPBasicBlock::SetLocalNames()

#if SMP_FULL_LIVENESS_ANALYSIS
// Create local def-use chains for all global names
// Important to remember that a chain can start without a def because the variable is Live-In;
//   in this case we give the chain a pseudo-def at address (first addr in block - 1) **!!**
// Variant on SSALocalRenumber
void SMPBasicBlock::CreateGlobalChains() {
	bool DebugFlag = false;
#if 0
	DebugFlag |= (0 == strcmp(".init_proc", this->MyFunc->GetFuncName()));
#endif
	size_t NumGlobals = this->MyFunc->NumGlobalNames();
	size_t LocalIndex;
	vector<int> LocalUseIndex;
	// Initialize LocalUseIndex and DUChain values for each local name.
	set<op_t, LessOp>::iterator NameIter;
	if (NumGlobals < 1)
		return;
	this->GlobalDUChains.ChainsByName.clear();
	this->GlobalDUChains.ChainsByName.reserve(NumGlobals);
	op_t TempOp;
	TempOp.type = o_void;
	SMPDUChainArray Temp(TempOp);
	if (DebugFlag)
		msg("Initializing GlobalChainsByName.\n");
	this->GlobalDUChains.ChainsByName.assign(NumGlobals, Temp);
	for (NameIter = this->MyFunc->GetFirstGlobalName(); NameIter != this->MyFunc->GetLastGlobalName(); ++NameIter) {
		LocalIndex = (size_t) ExtractGlobalIndex(*NameIter);
		LocalUseIndex.push_back(-1); // init LocalUse indices to -1; first DEF will make it 0
		// Set the local name for each DU chain array.
		if (LocalIndex > this->GlobalDUChains.ChainsByName.size())
			msg("LocalIndex %d out of bounds in CreateGlobalChains.\n", LocalIndex);
		if (DebugFlag)
			msg("Setting name for LocalIndex = %d\n", LocalIndex);
		this->GlobalDUChains.ChainsByName.at(LocalIndex).SetName(*NameIter);
	}

	// Iterate through all instructions in the block. For each DEF of a global name,
	//  set a new LocalUse index and start a new DU chain. For each USE of a global name,
	//  set the current LocalUse index for that name and add the instruction to the DU
	//  chain.
	list<list<SMPInstr>::iterator>::iterator InstIter;
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		size_t NameIndex;
		// USEs get referenced logically before DEFs, so start with them.
		set<DefOrUse, LessDefUse>::iterator CurrUse = (*InstIter)->GetFirstUse();
		while (CurrUse != (*InstIter)->GetLastUse()) {
			op_t UseOp = CurrUse->GetOp();
			set<op_t, LessOp>::iterator UseNameIter = this->MyFunc->FindGlobalName(UseOp);
			if (UseNameIter != this->MyFunc->GetLastGlobalName()) {
				// Found USE of a global name
				NameIndex = ExtractGlobalIndex(*UseNameIter);
				if (NameIndex > LocalUseIndex.size())
					msg("NameIndex %d out of range in CreateGlobalChains.\n", NameIndex);
				int LocalUseNum = LocalUseIndex.at(NameIndex);
				if (LocalUseNum == -1) { // LiveIn, use before def
					LocalUseIndex.at(NameIndex)++;
					LocalUseNum++;
					// Make the DEF be just before the start of the block
					SMPDefUseChain TempDefChain(UseOp, this->GetFirstAddr() - 1);
					this->GlobalDUChains.ChainsByName.at(NameIndex).DUChains.push_back(TempDefChain);
				}
				// Push address of USE instruction onto the DEF-USE chain for this LocalUseNum.
				this->GlobalDUChains.ChainsByName.at(NameIndex).DUChains.at(LocalUseNum).PushUse((*InstIter)->GetAddr());
			}
			++CurrUse;
		} // end for all USEs in the instruction.
		// Now do the DEFs.
		set<DefOrUse, LessDefUse>::iterator CurrDef = (*InstIter)->GetFirstDef();
		while (CurrDef != (*InstIter)->GetLastDef()) {
			size_t NameIndex;
			op_t DefOp = CurrDef->GetOp();
			set<op_t, LessOp>::iterator DefNameIter = this->MyFunc->FindGlobalName(DefOp);
			if (DefNameIter != this->MyFunc->GetLastGlobalName()) {
				// Found DEF of a global name.
				NameIndex = ExtractGlobalIndex(*DefNameIter);
				if (NameIndex > LocalUseIndex.size())
					msg("DEF NameIndex %d out of range in CreateGlobalChains.\n", NameIndex);
				++LocalUseIndex[NameIndex]; // move up to next LocalUse subscript number
#if 0
				// Before the push_back() call, we should have # chains equal to new LocalUseNum
				int LocalUseNum = LocalUseIndex.at(NameIndex);
				assert(LocalUseNum == this->GlobalDUChains.ChainsByName.at(NameIndex).DUChains.size());
#endif
				// Set up a new DU chain and push it onto the vector of DU chains for this
				//  local name. Push the DEF onto the list (it will be the first reference
				//  on the list, because the list was newly created).
				SMPDefUseChain TempDefChain(DefOp, (*InstIter)->GetAddr());
				this->GlobalDUChains.ChainsByName.at(NameIndex).DUChains.push_back(TempDefChain);
			}
			++CurrDef;
		} // end for all DEFs in the instruction
	} // end for all instructions in the block
	if (DebugFlag)
		msg("Exiting CreateGlobalChains()\n");
	return;
} // end of SMPBasicBlock::CreateGlobalChains()
#endif

// Create local DEF-USE chains and renumber all references to all names in LocalNames.
void SMPBasicBlock::SSALocalRenumber(void) {
	bool DebugFlag = false;
#if 0
	DebugFlag |= (0 == strcmp(".init_proc", this->MyFunc->GetFuncName()));
#endif
	size_t NumLocals = this->LocalNames.size();
	size_t LocalIndex;
	vector<int> SSAIndex;
	// Initialize SSAIndex and DUChain values for each local name.
	set<op_t, LessOp>::iterator NameIter;
	if (DebugFlag)
		msg("LocalNames size: %d\n", NumLocals);
	this->LocalDUChains.ChainsByName.clear();
	this->LocalDUChains.ChainsByName.reserve(NumLocals);
	op_t TempOp;
	TempOp.type = o_void;
	SMPDUChainArray Temp(TempOp);
	if (DebugFlag)
		msg("Initializing ChainsByName.\n");
	this->LocalDUChains.ChainsByName.assign(NumLocals, Temp);
	for (NameIter = this->LocalNames.begin(); NameIter != this->LocalNames.end(); ++NameIter) {
		LocalIndex = (size_t) ExtractGlobalIndex(*NameIter);
		SSAIndex.push_back(-1); // init SSA indices to -1; first DEF will make it 0
		// Set the local name for each DU chain array.
		if (LocalIndex > this->LocalDUChains.ChainsByName.size())
			msg("ERROR: LocalIndex %d out of bounds in SSALocalRenumber.\n", LocalIndex);
		if (DebugFlag)
			msg("Setting name for LocalIndex = %d\n", LocalIndex);
		this->LocalDUChains.ChainsByName.at(LocalIndex).SetName(*NameIter);
	}

	// Iterate through all instructions in the block. For each DEF of a local name,
	//  set a new SSA index and start a new DU chain. For each USE of a local name,
	//  set the current SSA index for that name and add the instruction to the DU
	//  chain.
	list<list<SMPInstr>::iterator>::iterator InstIter;
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		size_t NameIndex;
		// USEs get referenced logically before DEFs, so start with them.
		set<DefOrUse, LessDefUse>::iterator CurrUse = (*InstIter)->GetFirstUse();
		while (CurrUse != (*InstIter)->GetLastUse()) {
			op_t UseOp = CurrUse->GetOp();
			set<op_t, LessOp>::iterator UseNameIter = this->LocalNames.find(UseOp);
			if (UseNameIter != this->LocalNames.end()) {
				// Found USE of a local name.
				NameIndex = ExtractGlobalIndex(*UseNameIter);
				if (NameIndex > SSAIndex.size())
					msg("ERROR: NameIndex %d out of range in SSALocalRenumber.\n", NameIndex);
				int SSANum = SSAIndex.at(NameIndex);
				// Update the SSA subscript in the DEF-USE list for the instruction.
#if SMP_DEBUG_DATAFLOW
				if (DebugFlag) {
					msg("Setting SSANum to %d for local NameIndex %d\n", SSANum, NameIndex);
					msg("Addr %x USE operand: ", (*InstIter)->GetAddr()); 
					PrintOneOperand(UseOp, 0, -1);
					msg("\n");
				}
#endif
				CurrUse = (*InstIter)->SetUseSSA(UseOp, SSANum);
				if (SSANum >= 0) { // skip USE before DEF names
					if ((size_t) SSANum > this->LocalDUChains.ChainsByName.at(NameIndex).DUChains.size())
						msg("ERROR: SSANum %d out of range in SSALocalRenumber.\n", SSANum);
					// Push address of USE instruction onto the DEF-USE chain for this SSANum.
					this->LocalDUChains.ChainsByName.at(NameIndex).DUChains.at(SSANum).PushUse((*InstIter)->GetAddr());
				}
			}
			++CurrUse;
		} // end for all USEs in the instruction.
		// Now do the DEFs.
		set<DefOrUse, LessDefUse>::iterator CurrDef = (*InstIter)->GetFirstDef();
		while (CurrDef != (*InstIter)->GetLastDef()) {
			op_t DefOp = CurrDef->GetOp();
			set<op_t, LessOp>::iterator DefNameIter = this->LocalNames.find(DefOp);
			if (DefNameIter != this->LocalNames.end()) {
				// Found DEF of a local name.
				NameIndex = ExtractGlobalIndex(*DefNameIter);
				if (NameIndex > SSAIndex.size())
					msg("ERROR: DEF NameIndex %d out of range in SSALocalRenumber.\n", NameIndex);
				++SSAIndex[NameIndex]; // move up to next SSA subscript number
				int SSANum = SSAIndex.at(NameIndex);
				// Put new SSA number into DEF list entry in the instruction.
				CurrDef = (*InstIter)->SetDefSSA(DefOp, SSANum);
				// Before the push_back() call, we should have # chains equal to new SSANum
				assert(SSANum == (int) this->LocalDUChains.ChainsByName.at(NameIndex).DUChains.size());
				SetGlobalIndex(&DefOp, NameIndex);
				// Set up a new DU chain and push it onto the vector of DU chains for this
				//  local name. Push the DEF onto the list (it will be the first reference
				//  on the list, because the list was newly created).
				SMPDefUseChain TempDefChain(DefOp, (*InstIter)->GetAddr());
				this->LocalDUChains.ChainsByName.at(NameIndex).DUChains.push_back(TempDefChain);
			}
			++CurrDef;
		} // end for all DEFs in the instruction
	} // end for all instructions in the block
	if (DebugFlag)
		msg("Exiting SSALocalRenumber()\n");
	return;
} // end of SMPBasicBlock::SSALocalRenumber()

// For the newly defined type DefType for DefOp at instruction DefAddr, propagate the
//  type to all USEs in the local SSA chain for the DEF. If any USE types change,
// return true.
bool SMPBasicBlock::PropagateLocalDefType(op_t DefOp, SMPOperandType DefType, ea_t DefAddr, int SSANum) {
	bool changed = false;
	bool DebugFlag = false;
#if SMP_DEBUG_OPTIMIZATIONS
	DebugFlag |= (0 == strcmp("memset", this->MyFunc->GetFuncName()));
#endif
	set<op_t, LessOp>::iterator NameIter;

	if (DebugFlag) {
		msg("PropagateLocalDefType for DefType %d at %x ", DefType, DefAddr);
		PrintOperand(DefOp);
		msg("\n");
	}

	NameIter = this->LocalNames.find(DefOp);
	if (DebugFlag) msg("Returned from LocalNames.find()\n");
	if (NameIter == this->LocalNames.end()) {
		msg("ERROR: local DEF name at %x not in LocalNames: ", DefAddr);
		PrintOperand(DefOp);
		msg("\n");
	}
	else {
		// DefOp is in the LocalNames, as expected
		unsigned int RegIndex = ExtractGlobalIndex(*NameIter);
		list<list<SMPInstr>::iterator>::iterator InstIter;
		assert(0 <= SSANum);
		assert(SSANum < (int) this->LocalDUChains.ChainsByName.at(RegIndex).DUChains.size());
		ea_t StartAddr = this->LocalDUChains.ChainsByName.at(RegIndex).DUChains.at(SSANum).GetDef();
		assert(DefAddr == StartAddr);
		ea_t LastAddr = this->LocalDUChains.ChainsByName.at(RegIndex).DUChains.at(SSANum).GetLastUse();
		if (BADADDR == LastAddr)
			return false;  // DEF had no USEs, so no propagation
		set<DefOrUse, LessDefUse>::iterator CurrUse;
		// Go through the instructions in the block and find the instructions between StartAddr
		//  and LastAddr that have USEs of DefOp. These will all have SSANum, based on the
		//  local du-chain, so it does not have to be checked. If any USE in the chain has
		//  a type other than DefType, change it to DefType and set changed to true.
		for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
			ea_t CurrAddr = (*InstIter)->GetAddr();
			if (CurrAddr > LastAddr)
				break;
			if (CurrAddr > StartAddr) {
				CurrUse = (*InstIter)->FindUse(DefOp);
				if (CurrUse != (*InstIter)->GetLastUse()) { // found a USE in the chain
					if (CurrUse->GetType() == UNINIT) {
						assert(SSANum == CurrUse->GetSSANum());
						CurrUse = (*InstIter)->SetUseType(DefOp, DefType);
						assert(CurrUse != (*InstIter)->GetLastUse()); // found USE
						changed = true;
					}
					// **!!** Do we need to override any other types besides UNINIT?
				}
			}
		} // end for all instructions
	} // end if DefOp is not a local name .. error message ... else ... process

	return changed;
} // end of SMPBasicBlock::PropagateLocalDefType()

// Propagate the DefType of the DEF in DefOp with SSANum to all USEs. DefOp is a global
//  name, so the propagation will have to recurse into successor blocks, check their
//  phi functions for occurrences of DefOp, etc.
// Return true if any USEs have their types changed.
bool SMPBasicBlock::PropagateGlobalDefType(op_t DefOp, SMPOperandType DefType, int SSANum) {
	bool changed = false;
	bool FoundPhiDef = false;
	bool DebugFlag = false;
#if SMP_DEBUG_OPTIMIZATIONS
	DebugFlag |= (0 == strcmp("__strtod_internal", this->MyFunc->GetFuncName()));
#endif
	list<list<SMPInstr>::iterator>::iterator InstIter;
	set<DefOrUse, LessDefUse>::iterator CurrUse;
	set<DefOrUse, LessDefUse>::iterator CurrDef;
	SMPOperandType UseType;

	// Use the Processed flag to avoid infinite recursion.
	if (this->Processed)
		return false;
	else
		this->Processed = true;

	if (DebugFlag) {
		msg("PropagateGlobalDefType: DefType %d SSANum %d ", DefType, SSANum);
		PrintOperand(DefOp);
		msg("\n");
	}

	// This might be a recursive invocation, so the DefOp might be in the Phi functions.
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi != this->GetLastPhi()) { // Found a corresponding phi function
		if (DebugFlag) msg("Found Phi for DefOp\n");
		// If SSANum matches the DEF in the Phi function, then we are being
		//  called by InferPhiDefType() and we should proceed to process
		//  the instructions in the block.
		if (SSANum != CurrPhi->GetDefSSANum()) { // not DEF, so find USE
			if (DebugFlag) msg("Phi was a USE\n");
			// Now, find the SSANum entry in the phi arguments (USEs)
			bool FoundUse = false;
			size_t PhiIndex;
			for (PhiIndex = 0; PhiIndex < CurrPhi->GetPhiListSize(); ++PhiIndex) {
				if (SSANum == CurrPhi->GetUseSSANum(PhiIndex))  {
					FoundUse = true;
					if (DebugFlag) msg("Found the phi USE at index %d\n", PhiIndex);
					if (UNINIT == CurrPhi->GetUseType(PhiIndex)) {
						changed = this->SetPhiUseType(DefOp, PhiIndex, DefType);
						assert(changed);
						if (DebugFlag) msg("Changed the phi USE type\n");
					}
#if 1
					break; // **!!** COULD BE MULTIPLE USES HERE
#endif
				}
			}
			if (!FoundUse) {
				// Not actually a problem. E.g. if we find an inst:  add eax,4 that
				//  uses eax SSA #7 and DEFs eax SSA #8, we will call this function
				//  to propagate the type of SSA #8. There could be a phi function
				//  at the top of this block that DEFs eax SSA #7 from earlier USEs
				//  of eax. We will not find eax SSA #8 in that phi function, so we
				//  pass down to the code below to process all instructions.
				;
#if 0
				msg("ERROR: Failed to find Phi use of SSA %d in block %d for: ", SSANum,
					this->GetNumber());
				PrintOperand(DefOp);
				msg("\n");
#endif
			}
			else 
				return changed;  // end the recursion; phi function does not permit DefOp
								//   with SSANum to be referenced on this path again
		} // end if DefOp and SSANum are not the phi function DEF
		else {
			FoundPhiDef = true;
		}
	} // end if we found a phi function for DefOp

	// If not in the Phi functions, it could be in the instructions for the block.
	bool FoundUse = false;
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		CurrUse = (*InstIter)->FindUse(DefOp);
		if ((CurrUse == (*InstIter)->GetLastUse()) || (SSANum != CurrUse->GetSSANum())) {
			if (FoundUse) {
				// If we have already gone past the DEF and at least one use,
				//  another DEF will indicate that the original DEF-USE chain is
				//  finished. In that case, we can save a lot of time by not processing
				//  any more instructions and recursing into successor blocks, because
				//  the original DEF is now dead.
				CurrDef = (*InstIter)->FindDef(DefOp);
				if ((CurrDef != (*InstIter)->GetLastDef()) && (SSANum != CurrDef->GetSSANum())) {
					return changed;
				}
				else {
					continue;
				}
			}
			else {
				continue;
			}
		}
		// If we reach this point, we have found a use with the same SSA number.
		FoundUse = true;
		UseType = CurrUse->GetType();
		if (DebugFlag) {
			msg("Found USE with same SSANum at %x : %s\n", (*InstIter)->GetAddr(),
				(*InstIter)->GetDisasm());
		}
		if (UNINIT == UseType) {
			CurrUse = (*InstIter)->SetUseType(DefOp, DefType);
			assert(CurrUse != (*InstIter)->GetLastUse()); // found USE
			changed = true;
		}
		else if (DefType != UseType) {
			// Weird. USE already has a type, but it is not the DEF type.
			//  If we are refining a POINTER to a particular pointer type, fine.
			//  We don't want to change NUMERIC to something else, however,
			//  because it is possible to perform inherently NUMERIC operations
			//  on pointer types, e.g. def a pointer, use it as pointer, then
			//  perform hash function computations on it. Those computations
			//  (shift, xor, etc.) should truly remain NUMERIC.
			if (IsRefinedDataPtr(DefType) && (IsEqType(POINTER, UseType))) {
				msg("Refining POINTER to %d at %x %s for USE:", DefType,
					(*InstIter)->GetAddr(), (*InstIter)->GetDisasm());
				CurrUse->Dump();
				msg("\n");
				CurrUse = (*InstIter)->SetUseType(DefOp, DefType);
				assert(CurrUse != (*InstIter)->GetLastUse()); // found USE
				changed = true;
			}
			else {
				msg("WARNING: Not overriding type %d to %d at %x %s for USE:", UseType,
					DefType, (*InstIter)->GetAddr(), (*InstIter)->GetDisasm());
				CurrUse->Dump();
				msg("\n");
			}
		}
	} // end for all instructions in the block

	// We recurse into all successor blocks looking for more USEs. We know that there
	//  cannot be more USEs if the global name DefOp is not LiveOut from this block, so
	//  we can just return in that case.
	if (this->LiveOutSet.end() == LiveOutSet.find(DefOp))  // not LiveOut
		return changed;

	// Only recurse into successors that have DefOp in their LiveIn set. The others
	//  will not have Phi functions for DefOp, because we constructed fully pruned SSA
	//  form, nor will they have USEs of this SSANum.
	list<list<SMPBasicBlock>::iterator>::iterator SuccIter;
	for (SuccIter = this->GetFirstSucc(); SuccIter != this->GetLastSucc(); ++SuccIter) {
		if ((*SuccIter)->IsLiveIn(DefOp)) {
			changed |= (*SuccIter)->PropagateGlobalDefType(DefOp, DefType, SSANum);
		}
	}

	return changed;
} // end of SMPBasicBlock::PropagateGlobalDefType()

// Propagate the metadata Status for UseOp/SSANum to its local DEF.
// Return true if successful.
bool SMPBasicBlock::PropagateLocalMetadata(op_t UseOp, SMPMetadataType Status, int SSANum) {
	bool changed = false;
	unsigned int LocIndex;

	set<op_t, LessOp>::iterator NameIter;
	NameIter = this->FindLocalName(UseOp);
	if (this->GetLastLocalName() == NameIter) {
		return false;
	}
	LocIndex = ExtractGlobalIndex(*NameIter);

	assert(0 <= SSANum);
	assert(SSANum < (int) this->LocalDUChains.ChainsByName.at(LocIndex).DUChains.size());
	ea_t StartAddr = this->LocalDUChains.ChainsByName.at(LocIndex).DUChains.at(SSANum).GetDef();

	// Find the instruction at StartAddr
	list<list<SMPInstr>::iterator>::iterator InstIter;
	list<SMPInstr>::iterator CurrInst;
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		CurrInst = *InstIter;
		if (StartAddr == CurrInst->GetAddr()) {
			set<DefOrUse, LessDefUse>::iterator CurrDef, CurrUse, NextUse;
			CurrDef = CurrInst->FindDef(UseOp);
			assert(CurrDef != CurrInst->GetLastDef());
			assert(SSANum == CurrDef->GetSSANum());
			// Set the new metadata status.
			if (Status != CurrDef->GetMetadataStatus()) {
				CurrDef = CurrInst->SetDefMetadata(UseOp, Status);
				changed = (CurrDef != CurrInst->GetLastDef());

				// If source operand was memory, we have two cases.
				//  (1) The instruction could be a load, in which
				//  case we should simply terminate the
				//  propagation, because the prior DEF of a memory
				//  location is always considered live metadata
				//  already, and we do not want to propagate liveness
				//  to the address regs in the USE list.
				//  (2) We could have an arithmetic operation such
				//  as reg := reg arithop memsrc. In this case, we
				//  still do not want to propagate through the memsrc,
				//  but the register is both DEF and USE and we need
				//  to propagate through the register.
				if (CurrInst->HasSourceMemoryOperand()) {
					if (3 == CurrInst->GetOptType()) { // move inst
						break; // address regs are not live metadata
					}
					else if ((5 == CurrInst->GetOptType())
						|| (NN_and == CurrInst->GetCmd().itype)
						|| (NN_or == CurrInst->GetCmd().itype)
						|| (NN_xor == CurrInst->GetCmd().itype)) {
						// add, subtract, and, or with memsrc
						// Find the DEF reg in the USE list.
						CurrUse = CurrInst->FindUse(UseOp);
						assert(CurrUse != CurrInst->GetLastUse());
						changed |= this->PropagateLocalMetadata(UseOp,
							Status, CurrUse->GetSSANum());
						break;
					}
				} // end if memory source

				// Now, propagate the metadata status to all the
				//  non-memory, non-flags-reg, non-special-reg 
				//  (i.e. regular registers) USEs.
				CurrUse = CurrInst->GetFirstUse();
				while (CurrUse != CurrInst->GetLastUse()) {
					NextUse = CurrUse;
					++NextUse;
					op_t UseOp2 = CurrUse->GetOp();
					// NOTE: **!!** To be less conservative, we
					//  should propagate less for exchange category
					//  instructions.
					if ((UseOp2.type == o_reg) && (!UseOp2.is_reg(R_sp))
						&& (!(this->MyFunc->UsesFramePointer() && UseOp2.is_reg(R_bp)))
						&& (!UseOp2.is_reg(X86_FLAGS_REG))) {

						if (this->MyFunc->IsGlobalName(UseOp2)) {
							changed |= this->MyFunc->PropagateGlobalMetadata(UseOp2,
								Status, CurrUse->GetSSANum());
						}
						else {
							changed |= this->PropagateLocalMetadata(UseOp2,
								Status, CurrUse->GetSSANum());
						}
					}
					CurrUse = NextUse;
				} // end while all USEs
			} // end if Status != CurrInst status
			break;
		} // end if StartAddr is current inst addr
	} // end for all instructions

	return changed;
} // end of SMPBasicBlock::PropagateLocalMetadata()

// For the UNINIT type DEF DefOp at instruction DefAddr, see if all its USEs have
//  a single type. If so, set the DEF to that type and return true, else return false.
bool SMPBasicBlock::InferLocalDefType(op_t DefOp, unsigned int LocIndex, ea_t DefAddr) {
	bool changed = false;
	bool DebugFlag = false;
#if SMP_DEBUG_OPTIMIZATIONS
	DebugFlag |= (0 == strcmp("weightadj", this->MyFunc->GetFuncName()));
#endif

	if (DebugFlag) {
		msg("InferLocalDefType at %x ", DefAddr);
		PrintOperand(DefOp);
		msg("\n");
	}

	list<list<SMPInstr>::iterator>::iterator InstIter;
	list<list<SMPInstr>::iterator>::iterator DefInstIter;
	// Find the instruction at DefAddr
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		if (DefAddr > (*InstIter)->GetAddr()) {
			; // have not yet reached the DEF
		}
		else if (DefAddr == (*InstIter)->GetAddr()) {  // found DEF
			DefInstIter = InstIter;
			break;
		}
		else { // never found the DEF
			msg("ERROR: DefAddr %x not found in InferLocalDefType for DefOp: ", DefAddr);
			PrintOperand(DefOp);
			msg("\n");
			return false;
		}
	} // end for all instructions
	set<DefOrUse, LessDefUse>::iterator CurrDef;
	CurrDef = (*DefInstIter)->FindDef(DefOp);
	if (CurrDef == (*DefInstIter)->GetLastDef()) {
		msg("ERROR: DEF at %x not found in DEF list by InferLocalDefType for DefOp: ", DefAddr);
		PrintOperand(DefOp);
		msg("\n");
		return false;
	}
	int SSANum = CurrDef->GetSSANum();
	assert(0 <= SSANum);
	assert(SSANum < (int) this->LocalDUChains.ChainsByName.at(LocIndex).DUChains.size());
	ea_t StartAddr = this->LocalDUChains.ChainsByName.at(LocIndex).DUChains.at(SSANum).GetDef();
	assert(DefAddr == StartAddr);
	ea_t LastAddr = this->LocalDUChains.ChainsByName.at(LocIndex).DUChains.at(SSANum).GetLastUse();
	if (BADADDR == LastAddr)
		return false;  // DEF had no USEs, so no inference from USEs is possible
	set<DefOrUse, LessDefUse>::iterator CurrUse;
	// Go through the instructions in the block and find the instructions between StartAddr
	//  and LastAddr that have USEs of DefOp. These will all have SSANum, based on the
	//  local du-chain, so it does not have to be checked. If all USEs in the chain have
	//  a single type (other than UNINIT), change the DEF type to match the USE type
	//  and set changed to true.
	bool FirstUseSeen = false;
	bool FoundUninit = false;
	bool FoundNumeric = false;
	bool FoundPointer = false;
	bool FoundUnknown = false;
	SMPOperandType PtrType = UNINIT;
	SMPOperandType UseType = UNINIT;
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		ea_t CurrAddr = (*InstIter)->GetAddr();
		if (CurrAddr > LastAddr)
			break;
		if (CurrAddr > StartAddr) {
			CurrUse = (*InstIter)->FindUse(DefOp);
			if (CurrUse != (*InstIter)->GetLastUse()) { // found a USE in the chain
				if (!FirstUseSeen)
					FirstUseSeen = true;
				UseType = CurrUse->GetType();
				if (UNINIT == UseType)
					FoundUninit = true;
				else if (IsNumeric(UseType))
					FoundNumeric = true;
				else if (IsUnknown(UseType))
					FoundUnknown = true;
				else if (IsDataPtr(UseType)) {
					if (FoundPointer) {
						if (IsNotEqType(PtrType, UseType)) {
							msg("WARNING: Differing ptr types in local chain:");
							msg(" Prev: %d Current: %d %s\n", PtrType, UseType,
								(*InstIter)->GetDisasm());
						}
						PtrType = POINTER;
					}
					else {
						FoundPointer = true;
						PtrType = UseType;
					}
				}
			}
		}
	} // end for all instructions

	if (FirstUseSeen) {
		// See if we have a consistent type.
		// If we see any definite POINTER uses, we must set the DEF
		//  to type POINTER or a refinement of it.
		if (FoundPointer)
			UseType = PtrType;
		else if (FoundNumeric && !FoundUninit && !FoundUnknown)
			UseType = NUMERIC;
		else
			return false; // no POINTER, but no consistent type

		if (DebugFlag) msg("Inferring local DEF of type %d\n", UseType);
		CurrDef = (*DefInstIter)->SetDefType(DefOp, UseType);
		changed = true;
		if (FoundPointer && FoundUninit) {
			// We will propagate PtrType to the UNINIT elements of
			//  the DEF-USE chain.
			changed |= this->PropagateLocalDefType(DefOp, PtrType, DefAddr, CurrDef->GetSSANum());
		}
	}

	return changed;
} // end of SMPBasicBlock::InferLocalDefType()

// Iterate over all phi functions, calling InferPhiDefType() for any Phi DEF whose
//  type is still UNINIT.
// Return TRUE if any changes are made.
bool SMPBasicBlock::InferAllPhiDefTypes(void) {
	bool changed = false;
	bool NewChange;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi = this->GetFirstPhi();
	while (CurrPhi != this->GetLastPhi()) {
		if (UNINIT == CurrPhi->GetDefType()) {
			CurrPhi = this->InferPhiDefType(CurrPhi, NewChange);
			changed = (changed || NewChange);
			if (NewChange) {
				this->MyFunc->IncTypedPhiDefs();
			}
			else {
				this->MyFunc->IncUntypedPhiDefs();
			}
		}
		else {
			this->MyFunc->IncTypedPhiDefs();
		}
		++CurrPhi;
	}
	return changed;
}  // end of SMPBasicBlock::InferAllPhiDefTypes()

// If all USEs in a Phi function are set in agreement with each other, then infer
//   the Phi DEF type is the same as the USEs type, set it, propagate it,
//   and return true.
// Otherwise, if all USEs of the Phi DEF in the function have a common type,
//  infer the Phi DEF type from them and return true.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::InferPhiDefType(set<SMPPhiFunction, LessPhi>::iterator DefPhi, bool &changed) {
	changed = false;
	bool SkipToBackInference = false;
	bool PropagateChange = false;
	set<SMPPhiFunction, LessPhi>::iterator UsePhi;
	set<SMPPhiFunction, LessPhi>::iterator ReturnPhi = DefPhi;
	SMPOperandType MeetType = UNINIT;
	op_t DefOp = DefPhi->GetAnyOp();
	int SSANum = DefPhi->GetDefSSANum();
	bool SafeFunc = this->MyFunc->IsSafe();
	bool IndirectOp = MDIsIndirectMemoryOpnd(DefOp, this->MyFunc->UsesFramePointer());
	bool UnsafeOp = (IndirectOp || ((o_reg != DefOp.type) && (!SafeFunc)));

	assert(UNINIT == DefPhi->GetDefType());
	for (size_t index = 0; index < DefPhi->GetPhiListSize(); ++index) {
		if (IsEqType(UNINIT, DefPhi->GetUseType(index))) {
			SkipToBackInference = true;
			break;
		}
	}

	if (!SkipToBackInference) {
		// Must have seen all non-UNINIT types if we reach this point.
		MeetType = DefPhi->ConditionalMeetType();
		if (IsNotEqType(UNKNOWN, MeetType)) {
			// Don't infer UNKNOWN until conditional type propagation stage.
			ReturnPhi = this->SetPhiDefType(DefOp, MeetType);
			changed = true;
			PropagateChange = true;
		}
	}
	else if (!UnsafeOp) { // try back inference from USEs of Phi DEF
		SMPOperandType DefType;
		DefType = this->MyFunc->InferGlobalDefType(DefOp, SSANum, this, false);
		if (IsNotEqType(UNINIT, DefType)) {
			// Don't set changed, because there is no downward propagation
			//  to do. Instead, propagate only to UNINIT Phi USEs.
			size_t limit = DefPhi->GetPhiListSize();
			set<size_t> UseUpdates;
			UseUpdates.clear();
			for (size_t index = 0; index < limit; ++index) {
				if (UNINIT == DefPhi->GetUseType(index)) {
					UseUpdates.insert(index); // keep list needing update
				}
			}
			set<size_t>::iterator UpIter;
			for (UpIter = UseUpdates.begin(); UpIter != UseUpdates.end(); ++UpIter) {
				this->SetPhiUseType(DefOp, *UpIter, DefType);
				// NOTE: DefPhi iterator is stale after this update
			}
			ReturnPhi = this->SetPhiDefType(DefOp, DefType);
			// We don't set PropagateChange to true for the back inference case,
			//  because there is no propagating to do.
			changed = true;
		}
	}

	// Propagate down to USEs of this DEF.
	if (PropagateChange && !UnsafeOp) {
		this->MyFunc->ResetProcessedBlocks();
		changed |= this->PropagateGlobalDefType(DefOp, MeetType, SSANum);
	}
	return ReturnPhi;
} // end of SMPBasicBlock::InferPhiDefType()

#if SMP_FULL_LIVENESS_ANALYSIS
// Extension of IsRegDead for Global names.  Check and see if it overlaps with a DEF-USE chain
bool SMPBasicBlock::IsGlobalRegDead(ea_t InstAddr, op_t Operand, unsigned int RegIndex) const {
	bool FoundInLiveInSet = (LiveInSet.end() != LiveInSet.find(Operand));
	bool FoundInLiveOutSet = (LiveOutSet.end() != LiveOutSet.find(Operand));
	bool FoundInUpExposedSet = (UpExposedSet.end() != UpExposedSet.find(Operand));
	bool FoundInKillSet = (KillSet.end() != KillSet.find(Operand));
	bool HasLocalChains = (0 < this->GlobalDUChains.ChainsByName.at(RegIndex).DUChains.size());
	bool DebugFlag = (0x804837b == InstAddr);

	if (DebugFlag) {
		msg("IsGlobalRegDead for %x : ", InstAddr);
		PrintOperand(Operand);
		msg("\nLiveIn: %d LiveOut: %d UpExposed: %d VarKilled: %d\n", FoundInLiveInSet,
			FoundInLiveOutSet, FoundInUpExposedSet, FoundInKillSet);
	}

	size_t LocalUseIndex;
	size_t LocalUseMax = this->GlobalDUChains.ChainsByName.at(RegIndex).DUChains.size() - 1;
	if (DebugFlag) msg("A");

	// If there are no local chains for this global name, then it is either the pass-through
	//  case (LiveIn, LiveOut, !UpExposed, !VarKilled) or it must be dead.
	if (!HasLocalChains)
		return (!FoundInLiveInSet);

	if (DebugFlag) msg("A1");

	// If it is Live-In and Live-Out and not in the kill set,
	// it is live the whole time
	if (FoundInLiveInSet && FoundInLiveOutSet && !FoundInKillSet)
		return false;

	if (DebugFlag) msg("B");
	// If it is not Live-In and not in the Kill set it is dead the whole time
	if (!FoundInLiveInSet && !FoundInKillSet)
		return true;

	if (DebugFlag) msg("C");
	// If it is not in the UpExposed set and is in the Kill set, it is dead
	// before the first def
	if (!FoundInUpExposedSet && FoundInKillSet) {
		ea_t StartAddr = this->GlobalDUChains.ChainsByName.at(RegIndex).DUChains.at(0).GetDef();
		if ((BADADDR != StartAddr) && (InstAddr <= StartAddr))
			return true;
	}

	if (DebugFlag) msg("D");
	// If it is in the UpExposedSet, it is live until the last use of the first chain
	if (FoundInUpExposedSet) {
		ea_t LastAddr = this->GlobalDUChains.ChainsByName.at(RegIndex).DUChains.at(0).GetLastUse();
		if ((BADADDR != LastAddr) && (InstAddr <= LastAddr))
			return false;
	}

	if (DebugFlag) msg("E");
	// If it is not Live-Out, then it is dead after the last use
	if (!FoundInLiveOutSet) {
		ea_t LastAddr = this->GlobalDUChains.ChainsByName.at(RegIndex).DUChains.at(LocalUseMax).GetLastUse();
		if ((BADADDR != LastAddr) && (InstAddr > LastAddr))
			return true;
	}

	if (DebugFlag) msg("F");
	//If it is LiveOut and it is in the Kill Set, it will be live after the last definition	
	if (FoundInLiveOutSet && FoundInKillSet) {
		ea_t DefAddr = this->GlobalDUChains.ChainsByName.at(RegIndex).DUChains.at(LocalUseMax).GetDef();
		if ((BADADDR != DefAddr) && (DefAddr < InstAddr))  // **!!**  <= or < ?
			return false;
	}

	if (DebugFlag) msg("G");
	// See if any DEF-USE chains overlap InstAddr's USEs.
	for (LocalUseIndex = 0; LocalUseIndex <= LocalUseMax; ++LocalUseIndex) {
		// Need to check here to not break if this block does not have a def for this (LiveIn)
		// Skip this check if it is UpExposed and this is the first chain
		if (!FoundInUpExposedSet || LocalUseIndex > 0) {
			ea_t StartAddr = this->GlobalDUChains.ChainsByName.at(RegIndex).DUChains.at(LocalUseIndex).GetDef();
			if (StartAddr >= InstAddr)
				continue;  // cannot overlap USE if DEF is after InstAddr
		}
		ea_t LastAddr = this->GlobalDUChains.ChainsByName.at(RegIndex).DUChains.at(LocalUseIndex).GetLastUse();
		if (DebugFlag) msg(" LastAddr: %x\n", LastAddr);
		if ((BADADDR != LastAddr) && (LastAddr >= InstAddr)) {
			if (DebugFlag) msg("Ga");
			return false;
		}
	}

	if (DebugFlag) msg("H");
	return true; // no DEF-USE chains overlapped the instrumentation point
} // end of SMPBasicBlock::IsGlobalRegDead()
#endif

// If no DEF-USE chains overlap the instrumentation point for InstAddr (which logically
//  falls between the USEs of the instruction at InstAddr and its DEFs), for register RegIndex,
//  then it is safe for the instrumentation to use that register without saving and
//  restoring its value. Return true in this case, false if there is DEF-USE overlap.
bool SMPBasicBlock::IsRegDead(ea_t InstAddr, unsigned int RegIndex) const {
	size_t SSAIndex;
	// See if any DEF-USE chains overlap InstAddr's USEs.
	assert(RegIndex < this->LocalDUChains.ChainsByName.size());
	for (SSAIndex = 0; SSAIndex < this->LocalDUChains.ChainsByName.at(RegIndex).DUChains.size(); ++SSAIndex) {
		ea_t StartAddr = this->LocalDUChains.ChainsByName.at(RegIndex).DUChains.at(SSAIndex).GetDef();
		if (StartAddr >= InstAddr)
			continue;  // cannot overlap USE if DEF is after InstAddr
		ea_t LastAddr = this->LocalDUChains.ChainsByName.at(RegIndex).DUChains.at(SSAIndex).GetLastUse();
		if ((BADADDR != LastAddr) && (LastAddr >= InstAddr))
			return false;
	}
	return true; // no DEF-USE chains overlapped the instrumentation point
} // end of SMPBasicBlock::IsRegDead()

// For functions in which the flags register is a global name, determine whether the flags
//  are dead at the instruction located at addr. Do not call this function unless the
//  FlagOp is a global name.
bool SMPBasicBlock::AreGlobalFlagsDead(ea_t addr, op_t FlagOp) {
	if (this->FlagsDeadAfterLastUse && (addr > this->LastFlagsUse))
		return true;
	if (this->FlagsDeadBeforeFirstKill && (addr <= this->FirstFlagsDef))
		return true;
	return false;
} // end of SMPBasicBlock::AreGlobalFlagsDead()

// Mark the registers that are dead for each instruction in the block.
void SMPBasicBlock::MarkDeadRegs(void) {
	bool DebugFlag = false;
#if 0
	DebugFlag |= (0 == strcmp(".init_proc", this->MyFunc->GetFuncName()));
	DebugFlag |= true;
#endif
	set<op_t, LessOp>::iterator NameIter;
	set<op_t, LessOp>::iterator FlagNameIter;
	list<list<SMPInstr>::iterator>::iterator InstIter;
	op_t FlagsOp;
	FlagsOp.type = o_reg;
	FlagsOp.reg = X86_FLAGS_REG;
	FlagsOp.specflag1 = 0;
	FlagsOp.specflag2 = 0;
	char DeadString[MAXSTR];
	bool GlobalFlags = this->MyFunc->IsGlobalName(FlagsOp);
	if (GlobalFlags) {
		this->FlagsDeadAfterLastUse = (this->LiveOutSet.find(FlagsOp) == this->LiveOutSet.end());
		this->FlagsDeadBeforeFirstKill = ((this->UpExposedSet.find(FlagsOp) == this->UpExposedSet.end())
			&& (this->KillSet.find(FlagsOp) != this->KillSet.end()));
		FlagNameIter = this->MyFunc->FindGlobalName(FlagsOp);
	}
	else {
		FlagNameIter = this->LocalNames.find(FlagsOp);
	}
	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		ea_t InstAddr = (*InstIter)->GetAddr();
		DebugFlag = (0x804837b == InstAddr);
		if (DebugFlag) {
			(*InstIter)->PrintOperands();
			msg("\n");
		}
		DeadString[0] = '\0';
		// First, put EFLAGS at beginning of string if it is dead.
		if (!GlobalFlags) {
			if (FlagNameIter != this->LocalNames.end()) {
				// EFLAGS is in the LocalNames
				unsigned int RegIndex = ExtractGlobalIndex(*FlagNameIter);
				if (this->IsRegDead(InstAddr, RegIndex)) {
					qstrncat(DeadString, " EFLAGS", sizeof(DeadString) - 1);
				}
			}
			else { // EFLAGS are not locally used, but not global either; they are dead here
				qstrncat(DeadString, " EFLAGS", sizeof(DeadString) - 1);
			}
		}
		else {
#if SMP_FULL_LIVENESS_ANALYSIS
			assert(FlagNameIter != this->MyFunc->GetLastGlobalName());
			unsigned int RegIndex = ExtractGlobalIndex(*FlagNameIter);
			bool newCatch, oldCatch;
			newCatch = this->IsGlobalRegDead(InstAddr, FlagsOp, RegIndex);
			oldCatch = this->AreGlobalFlagsDead(InstAddr, FlagsOp);	
			if (newCatch || oldCatch)
				qstrncat(DeadString, " EFLAGS", sizeof(DeadString) - 1);
#if 0
			if (newCatch && !oldCatch)
				msg("We caught a new Flags!\n");
#endif
			if (!newCatch && oldCatch)
				msg("How did that slip by?\n");
#if 0
			if (newCatch && oldCatch)
				msg("Not a new success.\n");
#endif
#else
			if (this->AreGlobalFlagsDead(InstAddr, FlagsOp)) {
				qstrncat(DeadString, " EFLAGS", sizeof(DeadString) - 1);
			}
			else {
				;
#if SMP_DEBUG_OPTIMIZATIONS
				msg("Global EFLAGS in %s\n", this->MyFunc->GetFuncName());
#endif
			}
#endif   // end if SMP_FULL_LIVENESS_ANALYSIS ... else ...
		}

		// Now, process all other local regs and skip EFLAGS.
		for (NameIter = this->LocalNames.begin(); NameIter != this->LocalNames.end(); ++NameIter) {
			if (NameIter->type != o_reg)
				continue;
			// We never want to consider ESP to be available for instrumentation.
			if (NameIter->is_reg(R_sp))
				continue;
			// Until we analyze interprocedural register use, it is safest to not
			//  use EBP for instrumentation. This could change with more analysis.
			if (NameIter->is_reg(R_bp))
				continue;

			unsigned int RegIndex = ExtractGlobalIndex(*NameIter);
			int RegNum = NameIter->reg;
			if (RegNum == X86_FLAGS_REG) {
				//  We moved EFLAGS to the beginning, per annotations spec
				continue;
			}
			if (this->IsRegDead(InstAddr, RegIndex)) {
				qstrncat(DeadString, " ", sizeof(DeadString) - 1);
				qstrncat(DeadString, RegNames[RegNum], sizeof(DeadString) - 1);
			}
		} // end for all local names

#if SMP_FULL_LIVENESS_ANALYSIS
		// Now, process all other global regs and skip EFLAGS.
		//msg(" Done.  Analyzing global regs...");
		size_t NumGlobals = this->MyFunc->NumGlobalNames();
		if (NumGlobals > 0) {
			for (NameIter = this->MyFunc->GetFirstGlobalName(); NameIter != this->MyFunc->GetLastGlobalName(); ++NameIter) {
				op_t RegOp;
				RegOp.type = o_reg;
				//msg(" i ");

				//msg("%d", NameIter->type);
				if (NameIter->type != o_reg)
					continue;
				//msg(" j ");
				// We never want to consider ESP to be available for instrumentation.
				if (NameIter->is_reg(R_sp))
					continue;
				//msg(" k ");
				// Until we analyze interprocedural register use, it is safest to not
				//  use EBP for instrumentation. This could change with more analysis.
				if (NameIter->is_reg(R_bp))
					continue;

				//msg(" l ");
				unsigned int RegIndex = ExtractGlobalIndex(*NameIter);
				int RegNum = NameIter->reg;
				if (RegNum == X86_FLAGS_REG) {
					//  We moved EFLAGS to the beginning, per annotations spec
					continue;
				}
				//msg(" m ");
				RegOp.reg = NameIter->reg;
				//msg("%i", RegOp.reg);
				if (this->IsGlobalRegDead(InstAddr, RegOp, RegIndex)) {
					qstrncat(DeadString, " ", sizeof(DeadString) - 1);
					qstrncat(DeadString, RegNames[RegNum], sizeof(DeadString) - 1);
				}
			} // end for all global names
		}
#endif
		//msg("Done.  Next instruction.\n");

		if (strlen(DeadString) > 1) {
			(*InstIter)->SetDeadRegs(DeadString);
		}
	} // end for all instructions
	if (DebugFlag)
		msg("Exiting MarkDeadRegs()\n");
	return;
} // end of SMPBasicBlock::MarkDeadRegs()

// erase() block starting at FirstAddr from Preds list
void SMPBasicBlock::ErasePred(ea_t FirstAddr) {
	list<list<SMPBasicBlock>::iterator>::iterator PredIter;
	PredIter = this->Predecessors.begin();
	while (PredIter != this->Predecessors.end()) {
		if ((*PredIter)->GetFirstAddr() == FirstAddr) {
			PredIter = this->Predecessors.erase(PredIter);
			break; // can only appear once
		}
		else
			++PredIter;
	}
	return;
} // end of SMPBasicBlock::ErasePred()

// Erase a phi function from the list.
bool SMPBasicBlock::ErasePhi(op_t OldOp) {
	set<SMPPhiFunction, LessPhi>::iterator PhiIter;
	PhiIter = this->FindPhi(OldOp);
	if (PhiIter == this->PhiFunctions.end())
		return false; // did not find, cannot erase
	this->PhiFunctions.erase(PhiIter);
	return true;
} // end of SMPBasicBlock::ErasePhi()

// Add a phi function to the list of phi functions entering this block.
// If phi function for this global name already existed in the block,
//   return false because no new phi function was added; else return true.
bool SMPBasicBlock::AddPhi(SMPPhiFunction NewPhi) {
	if (this->PhiFunctions.end() == this->PhiFunctions.find(NewPhi)) {
		this->PhiFunctions.insert(NewPhi);
		return true;
	}
	else
		return false;
} // end of SMPBasicBlock::AddPhi()

// Change the SSA number of the Phi function DEF corresponding to DefOp.
bool SMPBasicBlock::SetPhiDefSSA(op_t DefOp, int SSANum) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		msg("ERROR: SetPhiDefSSA for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetSSADef(SSANum);
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		updated = this->AddPhi(TempPhi);
		assert(updated);
	}
	return updated;
} // end of SMPBasicBlock::SetPhiDefSSA()

// Change the SSA number of the Phi function USE corresponding to DefOp and index.
bool SMPBasicBlock::SetPhiUseSSA(op_t DefOp, size_t index, int SSANum) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		msg("ERROR: SetPhiUseSSA for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetSSARef(index, SSANum);
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		updated = this->AddPhi(TempPhi);
		assert(updated);
	}
	return updated;
} // end of SMPBasicBlock::SetPhiUseSSA()

// Change the SMP type of the Phi function DEF corresponding to DefOp.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::SetPhiDefType(op_t DefOp, SMPOperandType Type) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		msg("ERROR: SetPhiDefType for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetDefType(Type, & * this->Instrs.front());
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		pair<set<SMPPhiFunction, LessPhi>::iterator, bool> InsertResult;
		InsertResult = this->PhiFunctions.insert(TempPhi);
		CurrPhi = InsertResult.first;
		updated = InsertResult.second;
		assert(updated);
	}
	return CurrPhi;
} // end of SMPBasicBlock::SetPhiDefType()

// Change the SMP type of the Phi function USE corresponding to DefOp and index.
bool SMPBasicBlock::SetPhiUseType(op_t DefOp, size_t index, SMPOperandType Type) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		msg("ERROR: SetPhiUseType for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetRefType(index, Type,  & * this->Instrs.front());
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		updated = this->AddPhi(TempPhi);
		assert(updated);
	}
	return updated;
} // end of SMPBasicBlock::SetPhiUseType()

// Change the metadata type of the Phi function DEF corresponding to DefOp.
set<SMPPhiFunction, LessPhi>::iterator SMPBasicBlock::SetPhiDefMetadata(op_t DefOp, SMPMetadataType Status) {
	bool updated = false;
	set<SMPPhiFunction, LessPhi>::iterator CurrPhi;
	CurrPhi = this->FindPhi(DefOp);
	if (CurrPhi == this->PhiFunctions.end()) { // did not find
		msg("ERROR: SetPhiDefMetadata for block %d did not find DefOp: ", this->GetNumber());
		PrintOperand(DefOp);
		msg("\n");
	}
	else { // found it
		// To change a set item, we must change a copy, remove the original item
		//  from the set, and then insert the modified copy of the item.
		SMPPhiFunction TempPhi = (*CurrPhi);
		TempPhi.SetDefMetadata(Status);
		bool erased = this->ErasePhi(DefOp);
		assert(erased);
		pair<set<SMPPhiFunction, LessPhi>::iterator, bool> InsertResult;
		InsertResult = this->PhiFunctions.insert(TempPhi);
		CurrPhi = InsertResult.first;
		updated = InsertResult.second;
		assert(updated);
	}
	return CurrPhi;
} // end of SMPBasicBlock::SetPhiDefMetadata()

// Find consecutive DEFs of the same name within the block, both having
//  live metadata, that have the same type, making the second DEF
//  redundant metadata.
bool SMPBasicBlock::FindRedundantLocalMetadata(bool SafeFunc) {
	bool changed = false;
	bool DebugFlag = false;
	bool CallInst;
	bool UseFP = this->GetFunc()->UsesFramePointer();
	set<op_t, LessOp>::iterator NameIter;
	list<list<SMPInstr>::iterator>::iterator InstIter;
	list<SMPInstr>::iterator CurrInst;
	set<DefOrUse, LessDefUse>::iterator CurrDef;
	size_t LocalNameLimit = this->LocalDUChains.ChainsByName.size();
	size_t GlobalNameLimit = this->GlobalDUChains.ChainsByName.size();
	vector<SMPOperandType> LocalTypes;
	vector<SMPOperandType> GlobalTypes;

	if (0 == strcmp("__libc_start_main", this->GetFunc()->GetFuncName())) {
		DebugFlag = true;
	}

	// Initialize the most recently seen type for each possible DEF.
	for (size_t i = 0; i < LocalNameLimit; ++i)
		LocalTypes.push_back(UNINIT);
	for (size_t i = 0; i < GlobalNameLimit; ++i)
		GlobalTypes.push_back(UNINIT);

	for (InstIter = this->Instrs.begin(); InstIter != this->Instrs.end(); ++InstIter) {
		CurrInst = *InstIter;
		if (NN_fnop == CurrInst->GetCmd().itype)
			continue; // skip SSA marker instruction; has no normal DEFs
		CallInst = ((CurrInst->GetDataFlowType() == CALL)
			|| (CurrInst->GetDataFlowType() == INDIR_CALL));
		for (CurrDef = CurrInst->GetFirstDef(); CurrDef != CurrInst->GetLastDef(); ++CurrDef) {
			op_t DefOp = CurrDef->GetOp();
			// We limit our analysis to register DEFs, or stack memory DEFs
			//  in SAFE functions. **!!** Could add o_mem direct accesses
			//  in SAFE functions.
			if (((DefOp.type != o_reg) || DefOp.is_reg(R_sp) || (UseFP && DefOp.is_reg(R_bp)))
				&& (!(SafeFunc && MDIsStackAccessOpnd(DefOp, UseFP))))
				continue;
			// If the metadata is not live for the DEF it is irrelevant
			//  to redundancy analysis. We make an exception for the DEFs
			//  of caller-saved regs at a CALL instruction, because we have
			//  no idea what might have happened to the metadata in the callee.
			if ((CurrDef->GetMetadataStatus() != DEF_METADATA_USED)
				&& (!(CallInst && MDIsCallerSavedReg(DefOp))))
				continue;
			bool GlobalName = false;
			NameIter = this->GetFunc()->FindGlobalName(DefOp);
			unsigned int NameIndex;
			SMPOperandType CurrType = CurrDef->GetType();
			SMPOperandType OldType;
			if (NameIter != this->GetFunc()->GetLastGlobalName()) {
				// global name
				GlobalName = true;
				NameIndex = ExtractGlobalIndex(*NameIter);
				OldType = GlobalTypes.at(NameIndex); 
			}
			else { // local name
				NameIter = this->FindLocalName(DefOp);
				assert(NameIter != this->GetLastLocalName());
				NameIndex = ExtractGlobalIndex(*NameIter);
				OldType = LocalTypes.at(NameIndex); 
			}
			if (IsEqType(CurrType, OldType)) {
				// types agree; only redundant if NUMERIC or pointers with
				//  the same referent, which we cannot determine statically,
				//  so we stick to NUMERIC redundancy only.
				if (IsNumeric(CurrType)) {
					// Metadata is redundant; profiler-dependent or not?
					bool ProfDerived = IsProfDerived(CurrType)
						|| IsProfDerived(OldType);
					if (ProfDerived) {
						CurrDef = CurrInst->SetDefMetadata(DefOp, DEF_METADATA_PROF_REDUNDANT);
					}
					else {
						CurrDef = CurrInst->SetDefMetadata(DefOp, DEF_METADATA_REDUNDANT);
					}
					changed = true;
				}
			} // end if current type same as old type
			else {
				// Update latest type seen.
				if (GlobalName)
					GlobalTypes[NameIndex] = CurrType;
				else
					LocalTypes[NameIndex] = CurrType;
			}
		} // end for all DEFs in current instruction
	} // end for all instructions

	return changed;
} // end of SMPBasicBlock::FindRedundantLocalMetadata()
