#ifndef SMPINSTR_H
#define SMPINSTR_H 1

// SMPInstr.h
//
// This header defines the interfaces needed for analyzing instructions.

#include <cstddef>

#include <pro.h>
#include <ida.hpp>
#include <ua.hpp>

#include "SMPDataFlowAnalysis.h"

using namespace std;

enum SMPoperator {
	SMP_NULL_OPERATOR = 0,
	SMP_CALL,  // CALL instruction
	SMP_INPUT,  // input from port
	SMP_OUTPUT, // output to port
	SMP_ADDRESS_OF, // take effective address
	SMP_U_LEFT_SHIFT, // unsigned left shift
	SMP_S_LEFT_SHIFT, // signed left shift
	SMP_U_RIGHT_SHIFT, // unsigned right shift
	SMP_S_RIGHT_SHIFT, // signed right shift
	SMP_ROTATE_LEFT,
	SMP_ROTATE_LEFT_CARRY, // rotate left through carry
	SMP_ROTATE_RIGHT,
	SMP_ROTATE_RIGHT_CARRY, // rotate right through carry
	SMP_DECREMENT,
	SMP_INCREMENT,
	SMP_ADD,
	SMP_ADD_CARRY,   // add with carry
	SMP_SUBTRACT,
	SMP_SUBTRACT_BORROW,  // subtract with borrow
	SMP_U_MULTIPLY,
	SMP_S_MULTIPLY,
	SMP_U_DIVIDE,
	SMP_S_DIVIDE,
	SMP_U_REMAINDER,
	SMP_SIGN_EXTEND,
	SMP_ZERO_EXTEND,
	SMP_ASSIGN,
	SMP_BITWISE_AND,
	SMP_BITWISE_OR,
	SMP_BITWISE_NOT, // unary operator
	SMP_BITWISE_XOR,
	SMP_NEGATE,    // unary negation
	SMP_S_COMPARE, // signed compare (subtraction-based)
	SMP_U_COMPARE, // unsigned compare (AND-based)
	SMP_LESS_THAN, // boolean test operators
	SMP_GREATER_THAN,
	SMP_LESS_EQUAL,
	SMP_GREATER_EQUAL,
	SMP_EQUAL,
	SMP_NOT_EQUAL,
	SMP_LOGICAL_AND,
	SMP_LOGICAL_OR,
	SMP_UNARY_NUMERIC_OPERATION,  // miscellaneous; produces NUMERIC result
	SMP_BINARY_NUMERIC_OPERATION,  // miscellaneous; produces NUMERIC result
	SMP_SYSTEM_OPERATION,   // for instructions such as CPUID, RDTSC, etc.; NUMERIC
	SMP_UNARY_FLOATING_ARITHMETIC,  // all the same to our type system; all NUMERIC
	SMP_BINARY_FLOATING_ARITHMETIC  // all the same to our type system; all NUMERIC
}; // end enum SMPoperator
#define LAST_SMP_OPERATOR SMP_BINARY_FLOATING_ARITHMETIC

struct RTLoperator {
	SMPoperator oper;
	SMPOperandType type;
	SMPOperandType NonSpeculativeType;
};

// Class for a guard expression in an RTL
class SMPGuard {
public:
	// Constructors
	SMPGuard(void);
	// Get methods
	inline SMPoperator GetOperator(void) { return GuardOp; };
	inline op_t &GetLeftOperand(void) { return LeftOperand; };
	inline op_t &GetRightOperand(void) { return RightOperand; };
	// Set methods
	inline void SetOperator(SMPoperator op) { GuardOp = op; };
	inline void SetLeftOperand(op_t Left) { LeftOperand = Left; };
	inline void SetRightOperand(op_t Right) { RightOperand = Right; };
	// Printing methods
	void Dump(void);
private:
	op_t LeftOperand;
	op_t RightOperand;
	SMPoperator GuardOp;
}; // end class SMPGuard

// Classes for register transfers and register transfer lists (RTLs)
class SMPRegTransfer {
public:
	// Constructors and destructors
	SMPRegTransfer(void);
	~SMPRegTransfer();
	// Get methods
	inline SMPoperator GetOperator(void) { return RTop.oper; };
	inline SMPOperandType GetOperatorType(void) { return RTop.type; };
	inline op_t &GetLeftOperand(void) { return LeftOperand; };
	inline op_t &GetRightOperand(void) { return RightOperand; };
	inline SMPRegTransfer *GetRightTree(void) { return RightRT; };
	inline SMPGuard *GetGuard(void) { return Guard; };
	// Set methods
	inline void SetOperator(SMPoperator op) { RTop.oper = op; };
	void SetOperatorType(SMPOperandType OpType, const SMPInstr* instr);
	inline void SetLeftOperand(op_t Left) { LeftOperand = Left; };
	inline void SetRightOperand(op_t Right) { RightOperand = Right; };
	inline void SetRightTree(SMPRegTransfer *RightTree) { RightRT = RightTree; RightSubTree = true; };
	inline void SetGuard(SMPGuard *NewGuard) { Guard = NewGuard; };
	// Query methods
	inline bool HasRightSubTree(void) const { return RightSubTree; };
	// Printing methods
	void Dump(void);
private:
	SMPGuard *Guard;   // guard expression
	RTLoperator RTop;  // operator for the reg transfer
	op_t LeftOperand;  // operand to the left of the operator in the expression
	bool RightSubTree; // Is right operand another tree (true) or an operand (false)?
	op_t RightOperand; // valid only if RightSubTree is false
	SMPRegTransfer *RightRT; // right subtree, valid only if RightSubTree is true
}; // end class SMPRegTransfer

#define SMP_RT_LIMIT  16    // how many register transfers could be in an RTL?
class SMPRTL { // register transfer list
public:
	// Constructors and destructors.
	SMPRTL();
	~SMPRTL();
	// Get methods
	inline size_t GetCount(void) const { return RTCount; };
	SMPRegTransfer *GetRT(size_t index);
	// Set methods
	void push_back(SMPRegTransfer *NewEffect);
	// Printing methods
	void Dump(void);
	// Public data for ease of use.
	vector<op_t> ExtraKills;    // kills other than left operands
private:
	size_t RTCount;
	SMPRegTransfer *RTvector[SMP_RT_LIMIT];
}; // end class SMPRTL

class SMPInstr {
public: 
	// Constructors and destructors
	SMPInstr(ea_t addr);
	// Operators
	int operator==(const SMPInstr &rhs) const;
	int operator<(const SMPInstr &rhs) const;
	int operator<=(const SMPInstr &rhs) const;
	int operator!=(const SMPInstr &rhs) const;
    // Get methods
	inline ea_t GetAddr(void) const { return address; };
	inline char *GetDisasm(void) const { return (char *) disasm; };
	inline SMPBasicBlock *GetBlock(void) const { return BasicBlock; };
	inline set<DefOrUse, LessDefUse>::iterator GetFirstUse(void) { return Uses.GetFirstRef(); };
	inline set<DefOrUse, LessDefUse>::iterator GetFirstDef(void) { return Defs.GetFirstRef(); };
	inline set<DefOrUse, LessDefUse>::iterator GetLastUse(void)  { return Uses.GetLastRef(); };
	inline set<DefOrUse, LessDefUse>::iterator GetLastDef(void)  { return Defs.GetLastRef(); };
	inline set<DefOrUse, LessDefUse>::iterator FindUse(op_t SearchOp) { return Uses.FindRef(SearchOp); };
	inline set<DefOrUse, LessDefUse>::iterator FindDef(op_t SearchOp) { return Defs.FindRef(SearchOp); };
	inline size_t NumUses(void) const { return Uses.GetSize(); };
	inline size_t NumDefs(void) const { return Defs.GetSize(); };
	inline insn_t GetCmd(void) const { return SMPcmd; };
	inline int GetOptType(void) const { return OptType; };
	inline SMPitype GetDataFlowType(void) const { return type; };
	// Set methods
	inline void SetBlock(SMPBasicBlock *Block) { BasicBlock = Block; };
	inline void SetCmd(insn_t cmd) { SMPcmd = cmd; return; };
	inline void SetTerminatesBlock(void) { BlockTerm = true; };
	inline void SetTailCall(void) { 
		TailCall = true;
		CondTailCall = (type == COND_BRANCH);
		type = RETURN;
	};
	inline void AddDef(op_t DefOp, SMPOperandType DefType, int DefSSANum) {
		Defs.SetRef(DefOp, DefType, DefSSANum);
	}
	void MDAddRegDef(ushort DefReg, bool Shown, SMPOperandType Type = UNINIT);
		// Add DEF of register if not already a DEF
	void MDAddRegUse(ushort UseReg, bool Shown, SMPOperandType Type = UNINIT);
		// Add USE of register if not already a USE
	set<DefOrUse, LessDefUse>::iterator SetUseSSA(op_t CurrOp, int SSASub);
	set<DefOrUse, LessDefUse>::iterator SetDefSSA(op_t CurrOp, int SSASub);
	set<DefOrUse, LessDefUse>::iterator SetUseType(op_t CurrOp, SMPOperandType CurrType);
	set<DefOrUse, LessDefUse>::iterator SetDefType(op_t CurrOp, SMPOperandType CurrType);
	set<DefOrUse, LessDefUse>::iterator SetDefMetadata(op_t CurrOp, SMPMetadataType Status);
	inline void EraseUse(set<DefOrUse, LessDefUse>::iterator UseIter)  {
		Uses.EraseRef(UseIter);
		return;
	};
	inline void SetDeadRegs(char RegsString[]) { qstrncpy(DeadRegsString, RegsString, MAXSTR - 1); return; };
	// Query methods
	bool HasDestMemoryOperand(void); // Does instruction write to memory?
	bool HasSourceMemoryOperand(void); // Does instruction read from memory?
	bool IsSecondSrcOperandNumeric(flags_t F) const;
	bool IsBasicBlockTerminator(void) const;  // kind of inst that ALWAYS terminates a block
	inline bool IsLastInBlock(void) const { return BlockTerm; }; // does terminate its block
	inline bool IsJumpTarget(void) { return JumpTarget; };
	bool IsBranchToFarChunk(void);  // instr jumps outside current chunk
	inline bool IsTailCall(void) const { return TailCall; };
	inline bool IsCondTailCall(void) const { return CondTailCall; };
	inline bool MDIsInterruptCall(void) const { return Interrupt; };
	bool MDIsNop(void) const; // instruction is simple or complex no-op
	bool MDIsPushInstr(void) const;
	bool MDIsPopInstr(void) const;
	bool MDIsReturnInstr(void) const;
	bool MDIsEnterInstr(void) const;
	bool MDIsLeaveInstr(void) const;
	bool MDIsConditionalMoveInstr(void) const;
	bool MDIsStackPointerCopy(bool UseFP);  // copies ESP or EBP to register
	bool MDIsFrameAllocInstr(void);
	bool MDIsFrameDeallocInstr(bool UseFP, asize_t LocSize);
	bool MDUsesCalleeSavedReg(void);
	inline bool HasFlagsDef(void) const { return DefsFlags; };
	inline bool HasFlagsUse(void) const { return UsesFlags; };
	bool AllDEFsTyped(void); // No DEF is UNINIT
	bool AllUSEsTyped(void); // No USE is UNINIT
	// Printing methods
	void PrintOperands(void) const;
	char *DestString(int OptType);
	void Dump(void); // Complete debug print, with DEF/USE list, SSA #s, RTL
	// Analysis methods
	op_t GetSourceOnlyOperand(void); // return non-flags-reg non-dest source operand
	set<DefOrUse, LessDefUse>::iterator GetPointerAddressReg(op_t MemOp);
	void Analyze(void); // Fill in basic data for instruction.
	void AnalyzeMarker(void); // Fill in basic data for top of function pseudo-instruction.
	bool BuildRTL(void);   // Build RTL trees; return true if successfully built.
	void SyncAllRTs(void); // calls SyncRTLDefUse() for all RTs in RTL
	op_t GetPushedOpnd(void); // Extract source operand from PUSH RTL
	int MDGetImmedUse(void); // Get immed value from USE list of inst
	void SetImmedTypes(bool UseFP); // type all immediate operands as NUMERIC, CODEPTR, GLOBALPTR
									//  and set other context-free types (ESP == STACKPTR, etc.)
	bool InferTypes(void); // return true if any DEF-USE or RTL operator types were updated.
	void AnnotateStackConstants(bool UseFP, FILE *AnnotFile);
	void EmitAnnotations(bool UseFP, bool AllocSeen, FILE *AnnotFile); // No RTLs available
	void EmitTypeAnnotations(bool UseFP, bool AllocSeen, FILE *AnnotFile); // Use RTL types
	void EmitSafeReturn(FILE *AnnotFile); // emit annotation to denote that the return belongs to a safe function.
	void UpdateMemLoadTypes(SMPOperandType newType);

private:
	// Data 
	SMPBasicBlock *BasicBlock;  // basic block containing this instruction
	insn_t SMPcmd; // copy of 'cmd' for this instruction
	ulong features; // Canonical features for SMPcmd
	char disasm[MAXSTR]; // Disassembly text of instruction
	SMPitype  type; // Data flow analysis category
	int OptType;   // Optimization category (see OptCategory[])
	ea_t address;  // Code address for 1st byte of instruction
	bool analyzed; // Has instr been analyzed yet, setting type
	               //  and DEF and USE lists?
	bool JumpTarget; // Is Instr the target of any jumps or branches?
	bool BlockTerm;  // This instruction terminates a basic block.
	bool TailCall;  // This instruction is a tail call (jump to far chunk with stack restored).
	bool CondTailCall;  // Tail call is conditional branch.
	bool Interrupt;  // Instruction is a software interrupt call.
	bool RegClearIdiom; // ASM idiom for move zero into register?
	char DeadRegsString[MAXSTR]; // Registers that are dead at this instruction
	bool DefsFlags;  // Instr DEFs the flags
	bool UsesFlags;  // Instr USEs the flags
	SMPOperandType AddSubSourceType;  // Source op (USE only) type for add/sub
	op_t AddSubSourceOp; // operand corresponding to AddSubSourceType
	SMPOperandType AddSubUseType; // type of USE that is also DEFed by add /sub
	op_t AddSubUseOp; // operand corresponding to AddSubUseType
	op_t DestMemOp; // memory DEF operand, if any
	bool TypeInferenceComplete;  // All types have been resolved
	// For some type categories, inference just based on the category is done one time
	//  only, so this is a variable that is set to true after the first inference
	//  for those instruction types. For other categories, leave it false so that we
	//  try to infer types on each iteration of InferTypes().
	bool CategoryInferenceComplete;
	DefOrUseSet Defs; // Definitions list
	DefOrUseSet Uses; // Uses list
	SMPRTL RTL;

	// Methods
	bool AllDefsNumeric(void);  // true if all DEFs are NUMERIC or CODEPTR
	bool AllDefMetadataUnused(void); // true if all DEF metadata not needed
	void BuildSMPDefUseLists(void); // Build DEF and USE lists for instruction
	void MDFixupDefUseLists(void); // Machine-dependent ad hoc fixes
	void MDAnnotateSIBStackConstants(FILE *, op_t, ea_t, bool); // Handle x86 opcode SIB byte
	bool MDFindPointerUse(op_t MemOp, bool UseFP); // Set base reg to POINTER
	op_t MDGetMemUseOp(void);
	op_t MDGetMemDefOp(void);
	bool BuildUnaryRTL(SMPoperator UnaryOp); // helper for BuildRTL()
	bool BuildUnary2OpndRTL(SMPoperator UnaryOp); // helper for BuildRTL()
	bool BuildBinaryRTL(SMPoperator BinaryOp); // helper for BuildRTL()
	bool BuildLeaRTL(void); // helper for BuilRTL()
	bool BuildDoubleShiftRTL(SMPoperator BinaryOp); // helper for BuildRTL()
	bool BuildBinaryPlusFlagsRTL(SMPoperator BinaryOp); // helper for BuildRTL()
	bool BuildFlagsDestBinaryRTL(SMPoperator BinaryOp); // helper for BuildRTL()
	bool BuildCallRTL(void); // helper for BuildRTL()
	bool BuildReturnRTL(void); // helper for BuildRTL()
	bool BuildJumpRTL(SMPoperator CondBranchOp); // helper for BuildRTL()
	bool BuildEnterRTL(void); // helper for BuildRTL()
	bool BuildLeaveRTL(void); // helper for BuildRTL()
	bool BuildOptType8RTL(void); // helper for BuildRTL()
	bool BuildMoveRTL(SMPoperator GuardOp); // helper for BuildRTL()
	bool BuildCompareStringRTL(void); // helper for BuildRTL()
	bool BuildExchangeRTL(void); // helper for BuildRTL()
	bool BuildExchangeAddRTL(void); // helper for BuildRTL()
	bool BuildCompareExchangeRTL(void); // helper for BuildRTL()
	bool BuildPushRTL(void); // helper for BuildRTL()
	bool BuildPopRTL(void); // helper for BuildRTL()
	void AddToStackPointer(uval_t delta);
	void SubFromStackPointer(uval_t delta);
	bool BuildMultiplyDivideRTL(SMPoperator BinaryOp);  // helper for BuildRTL()
	void SyncRTLDefUse(SMPRegTransfer *CurrRT); // Ensure that all RTL operands are in the DEF/USE lists
	bool InferOperatorType(SMPRegTransfer *CurrRT); // return true if type updated
	void SetAddSubSourceType(void);
	SMPMetadataType GetDefMetadataType(void); // metadata type of non-flags DEF
};  // end class SMPInstr

#endif
